/* COPYRIGHT NOTICE
 *
 * This program comes from "The X Toolkit Cookbook" by Paul E. Kimball
 * c. 1995 Prentice Hall.
 *
 *  HDR1: Bonus Example - calculator.c - Four-function calculator
 *  
 *  HDR2: 
 *
 * This software is provided free of charge and may be freely distributed
 * in its original form, so long as this notice appears unchanged in each 
 * copy.  This software is supplied "as-is" and is not covered by warrantee
 * of any kind, express or implied.  
 *
 * This and other examples are available by anonymous ftp from
 * ftp.netcom.com, directory "pub/pk/pkimball/cookbook".
 *
 *  1/15/95	Paul Kimball - pkimball@netcom.com
 */

/***  Include Required Header Files   ********************************/

#include <strings.h>         /* string manipulation functions        */
#include <stdio.h>           /* standard IO functions                */
#include <X11/Intrinsic.h>   /* The Intrinsics                       */
#include <Xm/Xm.h>           /* Motif Symbols                        */
#include <Xm/PanedW.h>       /* Motif XmPanedWindow                  */
#include <Xm/RowColumn.h>    /* Motif XmRowColumn manager            */
#include <Xm/PushB.h>        /* Motif XmPushButton                   */

/***  Declare Global Variables  **************************************/

#define MAXINPUT  20              /* Max number of input digits      */
#define NONE       0              /* These are client_data that      */
#define ADD        1              /* are passed to callback          */
#define SUBTRACT   2              /* routines to tell them which     */
#define MULTIPLY   3              /* button was pressed and what     */
#define DIVIDE     4              /* to do.                          */
#define CLEARALL   6
#define EQUALS     7
                                  /* Flags to synchronize callbacks  */
Boolean       point_flag,         /* True if dec. point was pressed  */
              num_flag;           /* True if a number being entered  */
int           pending_operation;  /* pending operation               */
float         input, accum;       /* Accumulator, input              */
char          string[MAXINPUT+1], /* Input digits                    */
              character[2];       /* Latest digit                    */ 
int           num, i;             /* array counters                  */
Arg           arglist[10];        /* ArgList for SetValues           */
XtAppContext  context;            /* Application context             */
Widget        toplevel, calc_body;           /* Widget tree          */
Widget        button_panel, led_display;
Widget        add_button, subtract_button,   /* Function keys        */
              multiply_button, divide_button,
              point_button, equals_button,
              clearall_button, off_button;
Widget        number_button[10];             /* Number keys          */

/***  Forward Declare Callback Routines   ****************************/

void ExitRoutine();        /* Callback routine that exits            */
void FunctionKey();        /* Called when any function key is pressed*/
void NumberKey();          /* Called when any number key is pressed  */
void DecimalPoint();       /* Called when decimal point key pressed  */

/***  Main Program    ************************************************/

main(argc,argv)
int argc;
char **argv;
{

/*** Init. Toolkit, Appl. Context, Application Shell  ****************/

 toplevel = XtAppInitialize(&context, "Calculator", 
     NULL, 0, &argc, argv, NULL, NULL, 0);

/***  Create widget tree, bind callbacks  ****************************/

 calc_body = XtCreateManagedWidget(     /* overall manager           */
     "calcBody", xmPanedWindowWidgetClass, toplevel, NULL, 0);
 led_display = XtCreateManagedWidget(   /* XmPushButton display      */
     "ledDisplay",  xmPushButtonWidgetClass, calc_body, NULL, 0);
 button_panel = XtCreateManagedWidget(  /* XmRowColumn aligns buttons*/
     "buttonPanel", xmRowColumnWidgetClass, calc_body, NULL, 0);

/***  Create function key buttons; (all are similar) *****************/

 add_button = XtCreateManagedWidget(    /* The "ADD" XmPushButton    */
     "addButton", xmPushButtonWidgetClass, button_panel, NULL, 0);
 XtAddCallback(add_button, XmNactivateCallback,
     FunctionKey,                       /* function key callback     */
     (XtPointer)ADD );                  /* client_data identifies op */
 subtract_button = XtCreateManagedWidget("subtractButton",
     xmPushButtonWidgetClass, button_panel, NULL, 0);
 XtAddCallback( subtract_button, XmNactivateCallback,
     FunctionKey, (XtPointer)SUBTRACT );
 multiply_button = XtCreateManagedWidget("multiplyButton",
     xmPushButtonWidgetClass, button_panel, NULL, 0);
 XtAddCallback( multiply_button, XmNactivateCallback,
     FunctionKey, (XtPointer)MULTIPLY );
 divide_button = XtCreateManagedWidget("divideButton",
     xmPushButtonWidgetClass, button_panel, NULL, 0);
 XtAddCallback( divide_button, XmNactivateCallback,
     FunctionKey, (XtPointer)DIVIDE );

/***  Create number keys in a loop;  *********************************/

/*    Synthesize a unique name for each button, of the form          */
/*    "numberButtonN", where N is number represented by button       */

 for ( num=0; num<10; num++) {          /* Ten number keys           */
    sprintf(character, "%d", num);      /* Synthesize a unique name  */
    sprintf(string, "numberButton");    /* of the form numberButtonN */
    strcat(string, character);          /* add numeral to name       */
    number_button[num] = XtCreateManagedWidget( string,
        xmPushButtonWidgetClass, button_panel, NULL, 0);
    XtAddCallback( number_button[num], XmNactivateCallback,
        NumberKey,                      /* Callback routine          */
        (XtPointer)num );               /* Client_data is numeral    */
 }

/*** More Function keys **********************************************/

 point_button = XtCreateManagedWidget( "pointButton",
     xmPushButtonWidgetClass, button_panel, NULL, 0);
 XtAddCallback( point_button, XmNactivateCallback,
     DecimalPoint, (XtPointer)NULL );
 equals_button = XtCreateManagedWidget("equalsButton",
     xmPushButtonWidgetClass, button_panel, NULL, 0);
 XtAddCallback( equals_button, XmNactivateCallback,
     FunctionKey, (XtPointer)EQUALS );
 clearall_button = XtCreateManagedWidget("clearallButton",
     xmPushButtonWidgetClass, button_panel, NULL, 0);
 XtAddCallback( clearall_button, XmNactivateCallback,
     FunctionKey, (XtPointer)CLEARALL );
 off_button = XtCreateManagedWidget("offButton",
     xmPushButtonWidgetClass, button_panel, NULL, 0);
 XtAddCallback( off_button, XmNactivateCallback,
     ExitRoutine, (XtPointer)NULL );
 FunctionKey(NULL, CLEARALL, NULL);      /* Clear & init led display */
 XtRealizeWidget(toplevel);              /* Realize all              */
 XtAppMainLoop(context);                 /* Start main loop          */
}

/*** Callback Routines and Utilities  ********************************/

void ExitRoutine(w, client_data, call_data)    /* The friendly exit  */
Widget   w;                                    /* callback routine   */
XtPointer client_data, call_data;
{
  exit(1);
}
void DisplayLeds()                   /* Redisplays the "LED display" */
{
 XmString cstring;
 if (!num_flag)                      /* if function just performed   */
   sprintf(string, "%.2f", accum);   /* convert value in accumulator */
                                     /* put value in display         */
 cstring = XmStringCreateSimple(string); 
 XtVaSetValues(led_display, XmNlabelString, cstring, NULL);
 XmStringFree(cstring); 
}

/*********************************************************************/

void NumberKey(w, client_data, call_data)  /* Handle input           */
Widget w;                                  /* Widget calling back    */
XtPointer client_data;                     /* Input number           */
XtPointer call_data;                       /* unused                 */
{
 if (!num_flag){        /* first numeral in new input, clear display */
 num_flag = True;       /* Numeric input is now in progress          */
 string[0] = NULL;      /* Input string is zero length so far        */
 }
 sprintf(character, "%i", (int)client_data );  /* convert to string  */
 if(strlen(string)<MAXINPUT)                   /* only if it fits    */
 strcat(string, character);                    /* add to display     */
 DisplayLeds();                                /* refresh display    */
}

/*********************************************************************/

void DecimalPoint(w, client_data, call_data)   /* Handle decimal pt. */
Widget   w;
XtPointer client_data;
XtPointer call_data;
{
 if (point_flag) return;   /* Someone already pressed it so return   */
 if (!num_flag){           /* It's the first entry in a number       */
   num_flag = True;        /* Treat it just like any number          */
   string[0] = NULL;       /* Clear the input string                 */
   strcat(string,"0");     /* Put a zero in front just for looks     */
 }                         /* In any case, do the following:         */
 strcat(string, ".");      /* add it to the display                  */
 DisplayLeds();            /* display the new value                  */
 point_flag = True;        /* Set flag so we don't press it again    */
}

/*********************************************************************/

void FunctionKey(w, client_data, call_data)   /* Handle function key */
Widget w;                                     /* Widget calling back */
XtPointer client_data;                        /* Identifies function */
XtPointer call_data;                          /* unused              */
{
/* If this is a function key (say, the plus key), we must convert
 * any character data in the input buffer to a floating point 
 * number.  The trick is, in order to chain arithmetic, we then 
 * have to perform the operation that was requested *before* 
 * the last number got entered, not the one that just now
 * got requested.  The operation being requested now will get
 * performed the *next* time someone presses a function key.
 */
 int operation = (int)client_data;          /* cast client_data      */
 if (num_flag) {                            /* only if new input     */
   sscanf(string,"%f", &input);             /* convert input string  */ 
   switch (pending_operation){              /* perform pending op    */
     case NONE: accum = input; break;             /* No-op           */
     case ADD: accum = accum+input; break;        /* Addition        */
     case SUBTRACT: accum = accum-input; break;   /* Subtraction     */
  case MULTIPLY: accum = accum*input; break;      /* Multiplication  */
  case DIVIDE: accum = accum/input; break;        /* Division        */
     case EQUALS: accum=accum; break;             /* Equals          */
   }
 }
 pending_operation = operation;             /* New pending operation */
 if (operation == CLEARALL) {
   accum = 0.;                              /* Clear accumulator     */     
   string[0] = NULL; strcat(string,"0");    /* Clear input string    */
 }
 if ((operation == CLEARALL)||(operation == EQUALS)) {
   input = 0.;                              /* Clear input number    */
   pending_operation = NONE;                /* No pending operation  */
 }
 num_flag = False;           /* Ready to enter a new number          */
 point_flag = False;         /* User can press decimal point again   */
 DisplayLeds();              /* display the new value                */
}

/*********************************************************************/

